==> gm-1.6.3_Linux.tar.gz: gm-1.6.3_Linux/libgm/gm_blocking_receive_no_spin.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

#include "gm_call_trace.h"
#include "gm_cmp64.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

static gm_recv_event_t sleep_event;


GM_ENTRY_POINT
gm_recv_event_t *
gm_blocking_receive_no_spin (gm_port_t * p)
{
  gm_recv_event_t *ret;

  GM_CALLED_WITH_ARGS (("%p", p));

  ret = gm_receive (p);
  if (gm_ntoh_u8 (ret->recv.type) != GM_NO_RECV_EVENT)
    {
      GM_RETURN_PTR (ret);
    }

  /* For potential compatibility with a multithreaded environment,
     block repeatedly until we receive something.  When we sleep,
     ask the firmware to wake us when the slot we are sleeping on
     is filled. */

  ret = gm_receive (p);
  if (gm_ntoh_u8 (ret->recv.type) == GM_NO_RECV_EVENT)
    {
      struct gm_wake_request_send_event *s;
      struct gm_wake_request_send_event batch_write;
      unsigned int slot_num;

      if (!p->wake_scheduled)
	{
#if !GM_STRUCT_WRITE_COMBINING
#  error
#endif
	  slot_num = p->recv_queue_slot - p->recv_queue_start;
	  gm_assert (slot_num < 1<<16);
	  batch_write.recv_slot = gm_hton_u16 (slot_num);
	  batch_write.type = gm_hton_u8 ((gm_u8_t) GM_WAKE_REQUEST_EVENT);
	  s = GM_SEND_QUEUE_SLOT (p, 0, 0, wake_request);
	  GM_COPY_TO_IO_SPACE (*s, batch_write);
	  GM_FLUSH_SEND_EVENT (p, s);
	  
	  p->wake_scheduled = 1;
	}
      
      /* We could just go to sleep here until a wake indicates a
	 receive has happened, but we don't for backwards
	 compatibility.  Instead, we return a forged _GM_SLEEP_EVENT,
	 which the caller will pass to gm_unknown() where the sleep
	 will occur.  This allows multithreaded callers to release any
	 serialization mutex before sleeping. */

      sleep_event.recv.type = gm_hton_u8 (_GM_SLEEP_EVENT);
      ret = &sleep_event;
    }
  GM_RETURN_PTR (ret);
}

==> gm-2.0.8_Linux.tar.gz: gm-2.0.8_Linux/libgm/gm_blocking_receive_no_spin.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_blocking_receive_no_spin.c
*/

#include "gm_call_trace.h"
#include "gm_cmp64.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

static gm_recv_event_t sleep_event;

/**
   gm_blocking_receive_no_spin() behaves just like gm_blocking_receive(),
   only it sleeps the current thread immediately if no receive is
   pending.  It is well suited to applications with more than one
   CPU-intensive thread per processor.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_NO_RECV_EVENT
   @retval GM_WAKE_REQUEST_EVENT
   @retval GM_SLEEP_EVENT
 
   @param p \b (IN) The GM port in use.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_recv_event_t *
gm_blocking_receive_no_spin (gm_port_t * p)
{
  gm_recv_event_t *ret;

  GM_CALLED_WITH_ARGS (("%p", p));

  ret = gm_receive (p);
  if (gm_ntoh_u8 (ret->recv.type) != GM_NO_RECV_EVENT)
    {
      GM_RETURN_PTR (ret);
    }

  /* For potential compatibility with a multithreaded environment,
     block repeatedly until we receive something.  When we sleep,
     ask the firmware to wake us when the slot we are sleeping on
     is filled. */

  ret = gm_receive (p);
  if (gm_ntoh_u8 (ret->recv.type) == GM_NO_RECV_EVENT)
    {
      struct gm_wake_request_send_event *s;
      struct gm_wake_request_send_event batch_write;
      gm_offset_t slot_num;

      if (!p->wake_scheduled)
	{
#if !GM_STRUCT_WRITE_COMBINING
#  error
#endif
	  slot_num = p->recv_queue_slot - p->recv_queue_start;
	  gm_assert (slot_num < 1<<16);
	  batch_write.recv_slot = gm_hton_u16 ((gm_u16_t)slot_num);
	  batch_write.type = gm_hton_u8 ((gm_u8_t) GM_WAKE_REQUEST_EVENT);
	  s = GM_SEND_QUEUE_SLOT (p, 0, 0, wake_request);
	  GM_COPY_TO_IO_SPACE (*s, batch_write);
	  GM_FLUSH_SEND_EVENT (p, s);
	  
	  p->wake_scheduled = 1;
	}
      
      /* We could just go to sleep here until a wake indicates a
	 receive has happened, but we don't for backwards
	 compatibility.  Instead, we return a forged _GM_SLEEP_EVENT,
	 which the caller will pass to gm_unknown() where the sleep
	 will occur.  This allows multithreaded callers to release any
	 serialization mutex before sleeping. */

      sleep_event.recv.type = gm_hton_u8 (_GM_SLEEP_EVENT);
      ret = &sleep_event;
    }
  GM_RETURN_PTR (ret);
}
